/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

// piMC_ppu.c
// - MontoCarlo simulation to calcualte PI

#ifdef LIB_SPE1
	#include <libspe.h>
#endif

#ifdef LIB_SPE2
	#include <libspe2.h>
	#include <pthread.h>
#endif

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <libmisc.h>

// NOTE: The following three includes have an order dependency to ensure that Sobol control block is defined.
#include <mc_rand_sb.h>
#include "../piMC.h"
#include "../../sobol_init_30_40.h"

#define NUMSPUS 1

//#define ITERATIONS 25 //0xfff//0x0000ffff //0x0ffffffB// 25
//#define RUNS 4000       // number of random numbers (*4 for float) (*2 for double)  // 2000
//#define DIMENSION 3   // should only affect sobol random numbers

extern spe_program_handle_t calcPI_spu[NUMSPUS];

#ifdef LIB_SPE1
	speid_t aSPEID[NUMSPUS];
#endif

// Static allocation of memory aligned to 128 byte boundary
spu_rand_CB_t  spu_cb[NUMSPUS]  __attribute__ ((aligned(128)));	// control block


#ifdef LIB_SPE2
   
typedef struct thread_args {
	spe_context_ptr_t spe;
	void * argp;
	void * envp;
} thread_args_t;

void *spe_thread(void *arg){
	unsigned int runflags = 0;
	unsigned int entry = SPE_DEFAULT_ENTRY;
	thread_args_t * targ = (thread_args_t *)arg;
    // run SPE context
	if(spe_context_run(targ->spe, &entry, runflags, targ->argp, targ->envp, NULL) <0 ) {
		printf("** spe_context_run error, errno = %d\n", errno);
		exit(10);
	}
	pthread_exit(NULL);
}
#endif


// ******************************************************************************************************
int get_args(int argc, char** argv, UINT32 *u32Dim, UINT32 *u32Runs, UINT32 *u32Iter)
{
	int i;

	// argument checking and extraction
	if ( (argc!=7) && (argc!=8)){
		printf("\nUSAGE: %s -d <dimensions> -r <runs> -i <iterations> \n", argv[0]);
		printf(" note: dimension x runs < 7 000  and dimension x runs x iterations < 6 250 000\n");
		printf("if not program terminates \n");
		printf("Example: %s -d 4 -r 1000 -i 2000 \n", argv[0]);
		exit (0);
	}

	i=atoi(argv[2]);
	if (i < 0) 
	{
		printf("error: dimension must be > 0!\n");
		return 1;
	}
  
	i=atoi(argv[4]);
	if (i < 1) 
	{
		printf("error: runs must be > 0!\n");
		return 1;
	}

	i=atoi(argv[6]);
	if (i < 1)
	{
		printf("error: iterations must be > 0!\n");
		return 1;
	}
    // Start at i = 1 to skip the command name
	for (i = 1; i < argc; i++) {

		// Check for a switch (leading "-")
		if (argv[i][0] == '-') {

			// Use the next character to decide what to do
            switch (argv[i][1]) {
				case 'd':       *u32Dim = atoi(argv[++i]); break;
				case 'r':       *u32Runs = atoi(argv[++i]); break;
				case 'i':       *u32Iter = atoi(argv[++i]); break;
				default:        fprintf(stderr, "Unknown argument %s\n", argv[i]); return 1;
			}
		}
	}

	if (((*u32Dim) * (*u32Runs)) > 7000) {
		printf("\n dimension x runs must not larger than 7000 because of local store limitation-> exit \n");
		return 1;
	}

	if (((*u32Dim) * (*u32Runs) * (*u32Iter)) > 6250000) {
		printf("\n dimension x runs x iteration  must not larger than 6 250 000 because of main memory limitation-> exit \n");
		return 1;
	}

	return 0;
}




// ***************************************************************************
int main(int argc, char* argv[])
{
    int i=0;


    printf("sizeof spu_cb[NUMSPUS]=%d\n", sizeof(spu_cb[NUMSPUS]) );    


#ifdef LIB_SPE1
	INT32 au32status[NUMSPUS];
#endif
    char strRNG[8][12];

	unsigned int u32Iter;
	unsigned int u32Dim;
	unsigned int u32Runs;


#ifdef LIB_SPE2
	pthread_t pts[NUMSPUS];
	spe_context_ptr_t spe[NUMSPUS];
	thread_args_t targ1[NUMSPUS];
	int rc;
#endif

        if( 0 != get_args(argc, argv, &u32Dim, &u32Runs, &u32Iter) ) {
		return 1;
	}


	strcpy(strRNG[0],"scalar-SDK\0");       
	strcpy(strRNG[1],"vector-SDK\0");
	strcpy(strRNG[2],"HW\0");
	strcpy(strRNG[3],"KS\0");
	strcpy(strRNG[4],"SB\0");
	strcpy(strRNG[5],"MT\0");
	strcpy(strRNG[6],"n.a.\0");
	strcpy(strRNG[7],"n.a.\0");

// 	Set up control block (4)
//	spu_cb[i].u32Len  = u32BufferLen;
	spu_cb[i].u32Iter = u32Iter;
	spu_cb[i].u32Dim  = u32Dim;
	spu_cb[i].u32Runs = u32Runs;
	spu_cb[i].u32Index = i;
        strcpy(spu_cb[i].strRNG, strRNG[i]);
	rand_sobol_set_CB( &(spu_cb[i].sobolCB), 0);
	// <<


#ifdef LIB_SPE1
	aSPEID[0] = spe_create_thread(0, &calcPI_spu[0], NULL, 0, -1, 0);
//	spe_wait(aSPEID[0], &au32status[0] , 0);
#endif

#ifdef LIB_SPE2
	if( (spe[i] = spe_context_create(0, NULL)) == NULL){
		printf("** spe_context_create error, i=%d, errno = %d\n", i, errno);
		exit(2);
	}
 
	if( spe_program_load(spe[i], &calcPI_spu[i])!=0){
		printf("** spe_program_load error, i=%d, errno = %d\n", i, errno);
		exit(3);
	}
   
	targ1[i].spe = spe[i];
	targ1[i].argp = (void *)&spu_cb[i];
	targ1[i].envp = NULL;

	/* Create thread */
	if ((rc = pthread_create(&pts[i], NULL, spe_thread, (void *)&targ1[i])) != 0) {
		printf ("** pthread_create error, error=%d\n", rc);
		exit(6);
	}
#endif

	// **************************
	// Wait for threads to finish
        // **************************
        
#ifdef LIB_SPE1
	spe_wait(aSPEID[0], &au32status[0] , 0);
#endif


#ifdef LIB_SPE2
	// Wait for the SPE threads to terminate.
	if ((rc = pthread_join(pts[i], NULL)) != 0) {
		printf ("** pthread_join error, error=%d)\n", rc);
		exit (7);
	}

//	pthread_join(pts[i], NULL);
                
	// Destroy context
	if ((rc = spe_context_destroy (spe[i])) != 0){
		printf ("**spe_context_destroy error, i=%d, errno=%d \n", i, rc);
		exit (20);
	}
                
#endif


// >> print the results
	for ( i = 0; i < NUMSPUS; i++)
	{
		printf("\n results from spu-call %i is:",i);
		
		printf("\n PI = %f for RNG = %s\n",
					spu_cb[i].pi, strRNG[spu_cb[i].u32RNG]);
		printf("\n");
	}
	// <<








	return 0;
}
