/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

// new_rand_functions.h

#ifndef __NEW_RAND_FCTS_H__
#define __NEW_RAND_FCTS_H__

#ifdef __SPU__
#include <spu_intrinsics.h>
#endif
#include "rand_v.h"  

// *******************************************************************************
// FUNCTION
//	vector double _rand_0_to_1_d2(void)
//
// DESCRIPTION
//	_rand_0_to_1_vd2 generates a vector of random double precision floating-  
//	point values in the half closed range [0, 1.0). 
// ********************************************************************************

static __inline vector double _rand_0_to_1_d2() {
  vector unsigned int ival;

  vector double Fval;
  vector double OneV = (vector double) {1.0, 1.0};
  vector unsigned long long ExpVSel = (vector unsigned long long) {0xfff0000000000000ULL, 0xfff0000000000000ULL};

	/*    _rand_v generates a random vector word containing 4 31-bit 
	 *      integer random values.*/
	
  ival = (vector unsigned int)_rand_v();

  Fval = spu_sel((vector double) ival, OneV, ExpVSel);  // +1
  Fval = spu_sub(Fval, OneV);   // -1 

  return (Fval);
}

// *****************************************************************************************
// FUNCTION
//	vector double _rand_minus1_to_1_d2(void)
//
// DESCRIPTION
//	_rand_dbl_minus1_to_1_v generates a vector of random double precision floating-  
//	point values in the half closed range [-1.0, 1.0). 
// *****************************************************************************************

static __inline vector double _rand_minus1_to_1_d2(){
  vector unsigned int ival;

//  unsigned int *V1ptr, *V2ptr;
//  vector double *val;

  vector double Fval;
  vector double TwoV = (vector double) {2.0, 2.0};
  vector double ThreeV = (vector double) {3.0, 3.0};
  vector unsigned long long ExpVSel = (vector unsigned long long) {0xfff0000000000000ULL, 0xfff0000000000000ULL};

	/*    _rand_v generates a random vector word containing 4 31-bit 
	 *      integer random values.*/
	
  ival = (vector unsigned int)_rand_v();
 
  Fval = spu_sel((vector double) ival, TwoV, ExpVSel);
  Fval = spu_sub(Fval, ThreeV);

  return (Fval);
}
#endif
