/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdio.h>
#include <spu_mfcio.h>
#include <libmisc.h>
#include <math.h>
#include <string.h>

// Note: These three are order dependent to ensure the Sobol control block and local types are defined.
#include <mc_rand_sb.h>
#include "../myType.h"
#include "mc_sphere.h"


#define DMA_TAG_0 0
#define DMA_TAG_1 1


/***************************************************************
 * Module to to the Monte Carlo simulation
*************************************************************************************************/
#define vec(x,i) spu_extract(x,i)
#define print_f4(x) printf(" %f %f %f %f", vec(x,0), vec(x,1), vec(x,2), vec(x,3))
#define print_u4(x) printf(" %u %u %u %u", vec(x,0), vec(x,1), vec(x,2), vec(x,3))
#define print_x4(x) printf(" %x %x %x %x", vec(x,0), vec(x,1), vec(x,2), vec(x,3))

// ***************************************************************************
double sphereIntegration_0(UINT32 runs, float f_radius, UINT32 dim, vector float * pf_vecRN, spu_rand_CB_t  *ptrCB)
{

        const float f_Rsquare = f_radius *  f_radius;
        const float f_2R = 2.0 * f_radius;
        UINT8 u8Hint;
        UINT32 i, j, k1, N, nHit;
        double  sum;
        float r, x, *pf, *pfr;

        u8Hint = (runs > 0);  // predict branch (5)

        strcpy(ptrCB->strRNG,"sphereIntegration_0 scalar\0");

        nHit = 0;

        N = 0;
        pfr = (float *)pf_vecRN;
        for(k1=0; k1 < runs; k1++)
        {
            for(i=0; i<4; i++)
            {
                r = 0;
                pf = (pfr + i);

                // calculate the square of the L2 - norm for 4 elements in parallel (1)
                for( j = 0; j < dim; j++)
                {
                        x = *pf;
                        x     = f_2R*x - f_radius;      // scale from [0,1] to [-r, r]
                        r = x*x + r;
                        pf += 4;
                }

                // Hit Test
                if (r < f_Rsquare)
		    nHit++;
                N++;
            }
            pfr +=  4 * dim;
        }

        // execute branch with branch prediction (5)
        if(__builtin_expect (N != 0, u8Hint))
                sum = (double)nHit/(double)N;
        else
                sum = (double)0;

        return (double)sum;
}

// End of file
