/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdio.h>
#include <spu_mfcio.h>
#include <libmisc.h>
#include <math.h>
#include <string.h>

// Note: These three are order dependent to ensure the Sobol control block and local types are defined.
#include <mc_rand_sb.h>
#include "../myType.h"
#include "mc_sphere.h"


#define DMA_TAG_0 0
#define DMA_TAG_1 1


/***************************************************************
 * Module to to the Monte Carlo simulation
*************************************************************************************************/
#define vec(x,i) spu_extract(x,i)
#define print_f4(x) printf(" %f %f %f %f", vec(x,0), vec(x,1), vec(x,2), vec(x,3))
#define print_u4(x) printf(" %u %u %u %u", vec(x,0), vec(x,1), vec(x,2), vec(x,3))
#define print_x4(x) printf(" %x %x %x %x", vec(x,0), vec(x,1), vec(x,2), vec(x,3))

// ***************************************************************************
double sphereIntegration_3(UINT32 runs, float f_radius, UINT32 dim, vector float * pf_vecRN, spu_rand_CB_t  *ptrCB)
{

        const vector float    float_vec_0 = spu_splats((float) 0);
        const vector float    float_vec_1 = spu_splats((float) 1 * f_radius );
        const vector float    float_vec_2 = spu_splats((float) 2 * f_radius );
        const vector float    float_vecSphereR = spu_splats( f_radius );
        const vector float    float_vecSphereRsquare = spu_mul(float_vecSphereR, float_vecSphereR);

        UINT8 u8Hint;
        UINT32 j, k1, N, nHit, u32Step;
        double  sum;
        vector float r0, r1, x0, x1, xn0, xn1, *pf_vec0, *pf_vec1;
        vector UINT32 u32_vecHitResult, u32_vec_nHit, u32_vecHitResult0, u32_vec_nHit0,  u32_vecHitResult1, u32_vec_nHit1;

        u8Hint = (runs > 0);  // predict branch (5)

        strcpy(ptrCB->strRNG,"sphereIntegration_3 simd preload_2\0");

        u32_vec_nHit0 = spu_splats((UINT32) 0);
        u32_vec_nHit1 = u32_vec_nHit0;
	u32Step = dim << 1;

        N = 0;
        pf_vec0 = pf_vecRN;
        for(k1=0; k1 < (runs >> 1); k1++)
        {

                pf_vec1 = pf_vec0 + dim;

                r0 = float_vec_0;
                r1 = r0;

                x0 = *pf_vec0;
                x0 = spu_msub(float_vec_2, x0, float_vec_1);

                x1 = *pf_vec1;
                x1 = spu_msub(float_vec_2, x1, float_vec_1);

                // calculate the square of the L2 - norm for 4 elements in parallel (1)
                for( j = 1; j < dim; j++)
                {
                        r0     = spu_madd(x0, x0, r0);               // Euklid Norm
                        xn0 = *(pf_vec0+j);            // load one RN-vector into register for next cycle
                        x0     = spu_msub(float_vec_2, xn0, float_vec_1);      // scale from [0,1] to [-r, r]

                        r1     = spu_madd(x1, x1, r1);               // Euklid Norm
                        xn1 = *(pf_vec1+j);            // load one RN-vector into register for next cycle
                        x1     = spu_msub(float_vec_2, xn1, float_vec_1);      // scale from [0,1] to [-r, r]
                }
                pf_vec0 = pf_vec0 + u32Step;

                r0 = spu_madd(x0, x0, r0);                          // Euklid Norm
                r1 = spu_madd(x1, x1, r1);

                // Hit Test:  the spu_cmpabsgt replaces 4 ifs (2)
                u32_vecHitResult0 = (vector UINT32)spu_cmpabsgt(float_vecSphereRsquare, r0);
                u32_vecHitResult1 = (vector UINT32)spu_cmpabsgt(float_vecSphereRsquare, r1);

                // analyse hit result, track the sum of hits in each vector component (3)
                u32_vec_nHit0 = spu_sub(u32_vec_nHit0, u32_vecHitResult0);
                u32_vec_nHit1 = spu_sub(u32_vec_nHit1, u32_vecHitResult1);
                N+=8;   // increment by 4 because of 4 RNs per vector
        }
        // sum over all vector elements with compute complexity of O(log(n)) (4)
        u32_vec_nHit = spu_add(u32_vec_nHit0, u32_vec_nHit1);
        u32_vecHitResult = spu_rlqwbyte(u32_vec_nHit, 8);
        u32_vec_nHit = spu_add(u32_vec_nHit, u32_vecHitResult);
        u32_vecHitResult = spu_rlqwbyte(u32_vec_nHit, 4);
        u32_vec_nHit = spu_add(u32_vec_nHit, u32_vecHitResult);
        nHit = spu_extract(u32_vec_nHit,0);


        // execute branch with branch prediction (5)
        if(__builtin_expect (N != 0, u8Hint))
                sum = (double)nHit/(double)N;
        else
                sum = (double)0;
        return (double)sum;
}
// End of file
