/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdio.h>
#include <spu_mfcio.h>
#include <libmisc.h>
#include <math.h>
#include <string.h>

// Note: These three are order dependent to ensure the Sobol control block and local types are defined.
#include <mc_rand_sb.h>
#include "../myType.h"
#include "mc_sphere.h"


#define DMA_TAG_0 0
#define DMA_TAG_1 1


/***************************************************************
 * Module to to the Monte Carlo simulation
*************************************************************************************************/
#define vec(x,i) spu_extract(x,i)
#define print_f4(x) printf(" %f %f %f %f", vec(x,0), vec(x,1), vec(x,2), vec(x,3))
#define print_u4(x) printf(" %u %u %u %u", vec(x,0), vec(x,1), vec(x,2), vec(x,3))
#define print_x4(x) printf(" %x %x %x %x", vec(x,0), vec(x,1), vec(x,2), vec(x,3))

// ***************************************************************************
double sphereIntegration_4(UINT32 runs, float f_radius, UINT32 dim, vector float * pf_vecRN, spu_rand_CB_t  *ptrCB)
{

        const vector float    float_vec_0 = spu_splats((float) 0);
        const vector float    float_vec_1 = spu_splats((float) 1 * f_radius );
        const vector float    float_vec_2 = spu_splats((float) 2 * f_radius );
        const vector float    float_vecSphereR = spu_splats( f_radius );
        const vector float    float_vecSphereRsquare = spu_mul(float_vecSphereR, float_vecSphereR);

        UINT32 u32Hint;
        UINT32 j, k1, N, nHit, step;
        double  sum;
        vector float r0, r1, r2, r3, x0, x1, x2, x3, xn0, xn1, xn2, xn3;
        vector float *pf_vec0, *pf_vec1, *pf_vec2, *pf_vec3;
        vector UINT32 u32_vecHitResult, u32_vec_Hit, u32_vec_Hit0, u32_vec_Hit1;
        vector UINT32 u32_vecHitResult0, u32_vecHitResult1, u32_vecHitResult2, u32_vecHitResult3;
        vector UINT32 u32_vec_nHit0, u32_vec_nHit1, u32_vec_nHit2, u32_vec_nHit3;

        u32Hint = (runs > 0);  // predict branch (5)

        strcpy(ptrCB->strRNG,"sphereIntegration_4 simd preload_4\0");

        u32_vec_nHit0 = spu_splats((UINT32) 0);
        u32_vec_nHit1 = u32_vec_nHit0;
        u32_vec_nHit2 = u32_vec_nHit0;
        u32_vec_nHit3 = u32_vec_nHit0;

        N = 0;
        step = (dim << 2) ;
        pf_vec0 = pf_vecRN;

        for(k1=0; k1 < (runs >> 2); k1++)
        {
                pf_vec1 = pf_vec0 + dim;
                pf_vec2 = pf_vec1 + dim;
                pf_vec3 = pf_vec2 + dim;

                r0 = float_vec_0;
                r1 = r0;
                r2 = r0;
                r3 = r0;

                x0 = *pf_vec0;
                x1 = *pf_vec1;
                x2 = *pf_vec2;
                x3 = *pf_vec3;


                x0 = spu_msub(float_vec_2, x0, float_vec_1);
                x1 = spu_msub(float_vec_2, x1, float_vec_1);
                x2 = spu_msub(float_vec_2, x2, float_vec_1);
                x3 = spu_msub(float_vec_2, x3, float_vec_1);

                // calculate the square of the L2 - norm for 4 elements in parallel (1)
                for( j = 1; j < dim; j++)
                {
                        r0     = spu_madd(x0, x0, r0);
                        xn0 = *(pf_vec0+j); 
                        r1     = spu_madd(x1, x1, r1);
                        xn1 = *(pf_vec1+j);
                        r2     = spu_madd(x2, x2, r2);
                        xn2 = *(pf_vec2+j);
                        r3     = spu_madd(x3, x3, r3);
                        xn3 = *(pf_vec3+j);

                        x0     = spu_msub(float_vec_2, xn0, float_vec_1);
                        x1     = spu_msub(float_vec_2, xn1, float_vec_1);
                        x2     = spu_msub(float_vec_2, xn2, float_vec_1);
                        x3     = spu_msub(float_vec_2, xn3, float_vec_1);
                }
                pf_vec0 = pf_vec0 + step ;

                r0 = spu_madd(x0, x0, r0);
                r1 = spu_madd(x1, x1, r1);
                r2 = spu_madd(x2, x2, r2);
                r3 = spu_madd(x3, x3, r3);

                // Hit Test
                u32_vecHitResult0 = (vector UINT32)spu_cmpabsgt(float_vecSphereRsquare, r0);
                u32_vecHitResult1 = (vector UINT32)spu_cmpabsgt(float_vecSphereRsquare, r1);
                u32_vecHitResult2 = (vector UINT32)spu_cmpabsgt(float_vecSphereRsquare, r2);
                u32_vecHitResult3 = (vector UINT32)spu_cmpabsgt(float_vecSphereRsquare, r3);

                // analyse hit result, track the sum of hits in each vector component (3)
                u32_vec_nHit0 = spu_sub(u32_vec_nHit0, u32_vecHitResult0);
                u32_vec_nHit1 = spu_sub(u32_vec_nHit1, u32_vecHitResult1);
                u32_vec_nHit2 = spu_sub(u32_vec_nHit2, u32_vecHitResult2);
                u32_vec_nHit3 = spu_sub(u32_vec_nHit3, u32_vecHitResult3);

                N+=16;
        }
        // sum over all vector elements with compute complexity of O(log(n)) (4)
        u32_vec_Hit0  = spu_add(u32_vec_nHit0, u32_vec_nHit1);
        u32_vec_Hit1 = spu_add(u32_vec_nHit2, u32_vec_nHit3);
        u32_vec_Hit = spu_add(u32_vec_Hit0, u32_vec_Hit1);
        u32_vecHitResult = spu_rlqwbyte(u32_vec_Hit, 8);
        u32_vec_Hit = spu_add(u32_vec_Hit, u32_vecHitResult);
        u32_vecHitResult = spu_rlqwbyte(u32_vec_Hit, 4);
        u32_vec_Hit = spu_add(u32_vec_Hit, u32_vecHitResult);
        nHit = spu_extract(u32_vec_Hit,0);


        // execute branch with branch prediction (5)
        if(__builtin_expect (N != 0, u32Hint))
                sum = (double)nHit/(double)N;
        else
                sum = (double)0;
        return (double)sum;
}
// End of file
